"""
Avakas Built-In Base Project Flavor
"""

import re
import os

from git import Repo, Git

from avakas.errors import AvakasError
from avakas.avakas import Avakas, register_flavor
from avakas.utils import sort_versions


@register_flavor('git-native')
class AvakasGitNative(Avakas):
    """
    Avakas Git Native - using tags only
    Similar to thorscmversion
    """
    PROJECT_TYPE = 'git-native'

    @classmethod
    def guess_flavor(cls, directory):
        """
        Always return false as this is an explicitly called
        flavor.
        """
        # pylint: disable=unused-argument
        return False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.version_filename = kwargs['filename']
        self.repo = self.__load_git()

    def __load_git(self):
        """Initializes our local git workspace."""
        repo = Repo(self.directory, search_parent_directories=True)
        if not repo:
            raise AvakasError("Unable to find associated git repo for %s." %
                              self.directory)

        return repo

    def __git_push(self, tag):
        """Push git tag if remote exists"""
        opt = self.options
        if opt['remote'] not in [r.name for r in self.repo.remotes]:
            return

        if tag:
            remote = self.repo.remote(name=opt['remote'])
            resp = remote.push(tag)[0]

        if resp.flags & 1024 or resp.flags & 32 or resp.flags & 16:
            raise AvakasError("Unexpected git error: %s" % resp.summary)

    def __create_git_tag(self):
        """Creates a git tag"""
        return self.repo.create_tag(self.version)

    def __determine_bump(self):
        """Will go through the Git history until the last version bump
        and look for hints that we want to "automatically" bump
        our version"""
        self.repo = self.__load_git()
        vsn = None
        reg = re.compile(r'(\#|bump:|\[)(?P<bump>(patch|minor|major))(.*|\])',
                         re.MULTILINE)
        for commit in self.repo.iter_commits(self.options['branch']):
            # we go iterate back to the last time we bumped the version
            if commit.message.startswith('Version bumped to'):
                break

            res = reg.search(commit.message)
            if res:
                bump = res.group('bump')
                if not vsn:
                    vsn = bump
                elif vsn == 'patch' and bump == 'minor':
                    vsn = 'minor'
                elif vsn == 'patch' and bump == 'major':
                    vsn = 'major'
                elif vsn == 'minor' and bump == 'major':
                    vsn = 'major'
        return vsn

    def write_versionfile(self):
        """Write the version file"""
        path = os.path.join(self.directory, self.version_filename)
        version_file = open(path, 'w')
        version_file.write("%s\n" % self.version)
        version_file.close()

    def write_git(self):
        """Write data to git"""
        tag = None

        if not self.options['dry']:
            if not self._version.build:
                tag = self.__create_git_tag()
                self.__git_push(tag=tag)

    def bump(self,
             bump=None,
             prerelease=False,
             prerelease_prefix=None, build_date=None):
        """
        When using 'auto', flavor will attempt to determine whether or not
        the project needs to be bumped from git log history. If keywords are
        detected, project will update it's version to detected bump level and
        return True. If no keywords are detected, default_bump level will be
        used and True returned. If not, bump will return False.
        """
        if bump == 'auto':
            bump = self.__determine_bump()
            if bump is None and self.options['default_bump']:
                bump = self.options['default_bump']

        return super().bump(
            bump=bump,
            prerelease=prerelease,
            prerelease_prefix=prerelease_prefix,
            build_date=build_date)

    def read(self):
        git = Git(self.directory)
        out = git.tag(merged="HEAD", sort="-creatordate")
        tags = out.splitlines()
        tags = [t.strip(self.options['tag_prefix']) for t in tags]
        tags = sort_versions(tags)
        if len(tags) >= 2:
            latest_tag = tags[-1]
        elif len(tags) == 1:
            latest_tag = tags[0]
        else:
            raise AvakasError("No initial tag found!")

        self.version = latest_tag
        self.write_versionfile()

        return self.version

    def write(self):
        """
        Write version out to file
        """
        self.write_git()
        self.write_versionfile()



"""avakas

The avakas tool is meant as an interface around version
metadata for assorted flavours of software projects.

For more information see https://github.com/otakup0pe/avakas
"""

from __future__ import print_function

import os
import sys
import argparse

from git import Repo

from .avakas import detect_project_flavor
from .errors import AvakasError


def get_repo(directory):
    """Load the git repository."""
    return Repo(directory, search_parent_directories=True)


def git_rev(directory):
    """Returns the first eight characters of HEAD"""
    return str(get_repo(directory).head.commit)[0:8]


def add_metadata(project, buildmeta=False, **kwargs):
    """
    Add metadata for set/bump actions
    """
    directory = kwargs['directory'][0]

    git_str = str(git_rev(directory))
    if buildmeta:
        metadata = (git_str,)
        metadata += ci_build_meta()
        project.apply_metadata(*metadata)

    return project


def ci_build_meta():
    """Return any CI system specific build metadata"""
    meta = ()
    if 'BUILD_NUMBER' in os.environ:
        meta = (os.environ['BUILD_NUMBER'],)
    elif 'TRAVIS_BUILD_NUMBER' in os.environ:
        meta = (os.environ['TRAVIS_BUILD_NUMBER'],)
    elif 'CIRCLE_BUILD_NUM' in os.environ:
        meta = (os.environ['CIRCLE_BUILD_NUM'],)
    elif ('GITHUB_RUN_ID' in os.environ) and \
         ('GITHUB_RUN_NUMBER' in os.environ):
        meta = (os.environ['GITHUB_RUN_ID'], os.environ['GITHUB_RUN_NUMBER'],)
    return meta


def cli_show_version(**kwargs):
    """Show the current flavour specific version for a project."""
    project = detect_project_flavor(**kwargs)
    if not project.read():
        raise AvakasError('Unable to extract current version')

    print("%s" % str(project.version))


def cli_bump_version(
        level=None,
        prerelease=False,
        prerelease_date=False, **kwargs):
    """Bump the flavour specific version for a project."""
    project = detect_project_flavor(**kwargs)
    if not project.read():
        raise AvakasError('Unable to extract current version')
    old_version = project.version

    if not project.bump(
            bump=level[0],
            prerelease=prerelease,
            prerelease_prefix=kwargs['prerelease_prefix'],
            build_date=prerelease_date):
        sys.exit(0)
    project = add_metadata(project, **kwargs)
    project.write()

    print("Version updated from %s to %s" %
          (old_version, str(project.version)))


def cli_set_version(prerelease=False,
                    prerelease_date=False,
                    prerelease_prefix=None,
                    **kwargs):
    """Manually set the flavour specific version for a project."""

    version = kwargs['version'][0]
    project = detect_project_flavor(**kwargs)
    original_version = project.version_obj

    project.version = version
    if prerelease:
        prerelease_version = project.get_next_prerelease_version(
            starting_version=original_version,
            prefix=prerelease_prefix,
            new_version=project.version_obj
        )
        project.make_prerelease(prerelease_version, build_date=prerelease_date)
    project = add_metadata(project, **kwargs)
    project.write()

    print("Version set to %s" % version)


def parse_args(parser):
    """Parse our command line arguments."""

    bump_levels = ['patch', 'minor', 'major', 'auto']

    parser = argparse.ArgumentParser(prog="avakas",
                                     description='Process some integers.')

    subparsers = parser.add_subparsers(dest='operation')

    common = argparse.ArgumentParser(add_help=False)
    common.add_argument('--tag-prefix', dest='tag_prefix',
                        help='Prefix for version tag name',
                        default='')

    common.add_argument('--branch', dest='branch',
                        help='Branch to use when updating git',
                        default='master')

    common.add_argument('--remote', dest='remote',
                        help='Git remote',
                        default='origin')

    common.add_argument('--filename', dest='filename',
                        help='File name. Used for fallback versioning.',
                        default='version')
    common.add_argument('--flavor', dest='flavor',
                        help='Automation flavor for the project',
                        default='auto')
    common.add_argument('directory', nargs=1,
                        help='Directory of the project', default=os.getcwd())

    writable = argparse.ArgumentParser(add_help=False)
    writable.add_argument('--build-meta', dest='buildmeta',
                          help='Apply build-meta to version',
                          action='store_true',
                          default=False)
    writable.add_argument('--skip-dirty', dest='skipdirty',
                          help='Skip checking if local repo is dirty',
                          action='store_true',
                          default=False)
    writable.add_argument('--skip-commit-changes', dest='commitchanges',
                          help='Skip commiting generated version files',
                          action='store_false',
                          default=True)
    writable.add_argument('--with-hooks', dest='with_hooks',
                          help='Run git hooks', default=False)
    writable.add_argument('--dry-run',
                          dest='dry',
                          help='Will not push to git',
                          action='store_true')
    writable.add_argument('--prerelease',
                          dest='prerelease',
                          help='Will include prebuild information. If '
                          ' no other prebuild options are specified '
                          ' then it will simply use the build info in place.',
                          action='store_true')
    writable.add_argument('--prerelease-date',
                          dest='prerelease_date',
                          help='Include a string representation of the '
                          'current date, down to the second, as part '
                          'of the prebuild.',
                          action='store_true')
    writable.add_argument('--prerelease-prefix',
                          dest='prerelease_prefix',
                          help='Use the given string as a prebuild prefix',
                          default=None)

    set_p = subparsers.add_parser('set', parents=[common, writable])
    set_p.add_argument('version', nargs=1,
                       help='Desired version to set')

    bump_p = subparsers.add_parser('bump', parents=[common, writable])
    bump_p.add_argument('level', nargs=1, choices=bump_levels,
                        help='Level to bump at', default='auto')
    bump_p.add_argument('--default-bump', dest='default_bump',
                        choices=bump_levels, help='Level to bump at',
                        default=None)

    subparsers.add_parser('show', parents=[common])

    return parser.parse_args()


def main():
    """Dat entrypoint"""
    parser = argparse.ArgumentParser(prog="avakas")
    args = parse_args(parser)

    if args.operation is None:
        parser.print_help()
        sys.exit(0)

    directory = os.path.abspath(args.directory[0])

    if not os.path.exists(directory):
        raise AvakasError("Directory %s does not exist." % directory)

    try:
        if args.operation == 'bump':
            cli_bump_version(**vars(args))
        elif args.operation == 'show':
            cli_show_version(**vars(args))
        elif args.operation == 'set':
            cli_set_version(**vars(args))
        else:
            parser.print_help()
    except AvakasError as err:
        print("Problem: %s" % err.message, file=sys.stderr)
        sys.exit(1)

    sys.exit(0)



"""
Avakas Built-In Erlang Project Flavor
"""

from glob import glob
from erl_terms import decode as erl_decode

from avakas.flavors.base import AvakasLegacy
from avakas.avakas import register_flavor
from avakas.errors import AvakasError
from avakas.utils import match_and_rewrite_lines


@register_flavor('erlang')
class AvakasErlangProject(AvakasLegacy):
    """
    Erlang Avakas Project Flavor
    """
    PROJECT_TYPE = 'erlang'

    @classmethod
    def guess_flavor(cls, directory):
        return len(glob("%s/src/*.app.src" % directory)) == 1

    def read(self):
        app_file = glob("%s/src/*.app.src" % self.directory)[0]
        version_handle = open(app_file, 'r')
        erl_terms = erl_decode(version_handle.read())
        version_handle.close()
        app_config = erl_terms[0][2]
        erlang_version = None
        for config in app_config:
            if config[0] == 'vsn':
                erlang_version = config[1]

        if not erlang_version:
            raise AvakasError('Unable to determine Erlang version')

        self.version = erlang_version
        return erlang_version

    def write(self):
        app_file = glob("%s/src/*.app.src" % self.directory)[0]
        app_handle = open(app_file, 'r')
        lines, updated = match_and_rewrite_lines(r'(.+vsn.+")(.+)(".+)',
                                                 app_handle, self.version)
        app_handle.close()
        if not updated:
            raise AvakasError('Unable to save Erlang version')

        app_handle = open(app_file, 'w')
        app_handle.write(''.join(lines))
        app_handle.close()



"""
Avakas Utility Functions
"""

import re
import sys
import os
from functools import cmp_to_key
import contextlib

from semantic_version import compare


@contextlib.contextmanager
def stdout_redirect():
    """ Forcefully redirect stdout to stderr """
    # http://marc-abramowitz.com/archives/2013/07/19/python-context-manager-for-redirected-stdout-and-stderr/
    try:
        oldstdchannel = os.dup(sys.stdout.fileno())
        os.dup2(sys.stderr.fileno(), sys.stdout.fileno())

        yield
    finally:
        if oldstdchannel is not None:
            os.dup2(oldstdchannel, sys.stdout.fileno())


def match_and_rewrite_lines(pattern, file_body, version):
    """
    Replace lines by regex
    """
    lines = []
    updated = False
    for line in file_body.splitlines():
        re_out = re.sub(pattern, r'\1%s\3', line)
        if re_out != line:
            updated = True
            lines.append(re_out % version)
        else:
            lines.append(line)

    return ('\n'.join(lines), updated)


def sort_versions(versions):
    """
    Sort a list of version strings by semantic version
    """
    return sorted(versions, key=cmp_to_key(compare))



"""
avakas main load
"""

import avakas.flavors

from .errors import AvakasError
from .avakas import Avakas, register_flavor

__all__ = [
    'Avakas',
    'AvakasError',
    'register_flavor',
    'flavors',
]



"""
Avakas Built-In Nodejs Project Flavor
"""

import json
import os

from avakas.flavors.base import AvakasLegacy
from avakas.avakas import register_flavor


@register_flavor('node')
class AvakasNodeProject(AvakasLegacy):
    """
    Nodejs Avakas Project Flavor
    """
    PROJECT_TYPE = 'node'

    @classmethod
    def guess_flavor(cls, directory):
        return os.path.exists("%s/package.json" % directory)

    def __read_package_json(self):
        manifest = os.path.join(self.directory, 'package.json')
        manifest_file = open(manifest, 'r')
        manifest_json = json.load(manifest_file)
        manifest_file.close()

        return manifest_json

    def __write_package_json(self, manifest_dict):
        manifest = os.path.join(self.directory, 'package.json')
        manifest_file = open(manifest, 'w')
        json.dump(manifest_dict,
                  manifest_file,
                  indent=4,
                  separators=(',', ': '),
                  sort_keys=True)
        manifest_file.close()

    @classmethod
    def __extract_version(cls, manifest_json):
        return manifest_json['version']

    def read(self):
        manifest = self.__read_package_json()
        self.version = self.__extract_version(manifest)

    def write(self):
        manifest = self.__read_package_json()
        manifest['version'] = self.version
        self.__write_package_json(manifest)



"""
Avakas Built-In Ansible Project Flavor
"""

import os

from git import Git

from avakas.flavors.base import AvakasLegacy
from avakas.avakas import register_flavor
from avakas.errors import AvakasError
from avakas.utils import sort_versions


@register_flavor('ansible')
class AvakasAnsibleProject(AvakasLegacy):
    """
    Ansible Avakas Project
    """
    PROJECT_TYPE = 'ansible'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        tag_prefix = self.options.get('tag_prefix', '')
        if tag_prefix not in ['v', '']:
            raise AvakasError('Cannot specify a tag prefix '
                              'with an Ansible Role')
        self.options['tag_prefix'] = 'v'

    @classmethod
    def guess_flavor(cls, directory):
        return os.path.exists("%s/meta/main.yml" % directory)

    def read(self):
        git = Git(self.directory)
        out = git.tag(merged="HEAD", sort="-creatordate")
        tags = out.splitlines()
        tags = [t.strip(self.options['tag_prefix']) for t in tags]
        tags = sort_versions(tags)
        latest_tag = tags[-1]

        self.version = latest_tag

        return self.version



"""
Avakas Built-In Project Flavors
"""

from .base import AvakasLegacy
from .ansible import AvakasAnsibleProject
from .chef import AvakasChefProject
from .erlang import AvakasErlangProject
from .git import AvakasGitNative
from .node import AvakasNodeProject

__all__ = [
    'AvakasAnsibleProject',
    'AvakasChefProject',
    'AvakasErlangProject',
    'AvakasGitNative',
    'AvakasLegacy',
    'AvakasNodeProject',
]



"""
Avakas classes and plugin handlers
"""

import copy
import datetime

from semantic_version import Version

from .errors import AvakasError


def detect_project_flavor(**kwargs):
    """
    Determines the project flavor for a given directory
    """

    flavor = kwargs.get('flavor', 'auto')

    if flavor == 'auto':
        matched = [f for n, f
                   in Avakas.project_flavors.items()
                   if f.guess_flavor(directory=kwargs['directory'][0])]
        if len(matched) == 1:
            project = matched[0]
        elif len(matched) == 0:
            project = Avakas.project_flavors['legacy']
        else:
            matched_names = [f.PROJECT_TYPE for f in matched]
            raise AvakasError("Multiple project flavor matches: %s" %
                              ", ".join(matched_names))
    else:
        if flavor in Avakas.project_flavors:
            project = Avakas.project_flavors[flavor]
        else:
            raise AvakasError('Unable to find flavor "%s"' % flavor)

    return project(**kwargs)


class Avakas():
    """
    Main instance of Avakas associated to a project and it's version
    """
    project_flavors = {}

    def __init__(self, **kwargs):
        self._version = Version('0.0.0')
        self.directory = kwargs['directory'][0]
        self.options = kwargs

    @property
    def version(self):
        """Get version"""

        tag_prefix = self.options.get('tag_prefix', '')
        return "%s%s" % (tag_prefix, self._version)

    @version.setter
    def version(self, version):
        """Set version"""
        if not isinstance(version, str):
            raise TypeError("version must be type of str")

        tag_prefix = self.options.get('tag_prefix', None)
        version = version.strip(tag_prefix)

        try:
            self._version = Version(version)
        except ValueError as err:
            raise AvakasError("Invalid version string %s" %
                              version) from err

    @property
    def version_obj(self):
        """
        Get a copy (not the original) of the `semantic_version.Version object
        which this instance uses to internally manage its version

        Returns:
            * `semantic_version.Version` : A copy of `self._version`
        """

        return copy.deepcopy(self._version)

    @classmethod
    def read(cls):
        """Read version data from a project"""
        return True

    @classmethod
    def write(cls):
        """Write version data to a project"""

    def bump(self,
             bump=None,
             prerelease=False,
             prerelease_prefix=None,
             build_date=None):
        """Bump version"""

        original = self._version

        if not bump:
            return False

        if bump == 'patch':
            self._version = self._version.next_patch()
        elif bump == 'minor':
            self._version = self._version.next_minor()
        elif bump == 'major':
            self._version = self._version.next_major()
        else:
            raise AvakasError("Invalid version component")

        if prerelease:
            prerelease_version = self.get_next_prerelease_version(
                original, prefix=prerelease_prefix
            )

            self.make_prerelease(
                prerelease_version,
                prefix=prerelease_prefix,
                build_date=build_date)

        if self._version == original:
            msg = "Attempted to set the version to its previous value!"
            raise AvakasError(msg)

        return True

    def get_next_prerelease_version(
            self, starting_version=None, prefix=None,
            new_version=None):
        """
        Return an integer representing the version of a given prerelase label
        (i.e. alpha, beta, rc) which comes next.

        if `starting_version` is not passed in, this will use `self`'s
        version object as the starting version.
        """

        if starting_version is None:
            starting_version = self._version

        if new_version is None:
            new_version = self._version

        # This \/ checks whether last version was a pre-release, and then
        # whether the beginning (at minimum) of the current pre-release
        # prefix (PRP) matches the intended PRP. This could resolve to
        # being `True` in the case that we're doing some sort of pre-pre
        # release, e.g. rc.1.dev.1 would match an attempted 'rc' bump), but
        # that's actually desirable (because bumping to the next 'rc'
        # should treat `dev.1` as if it doesn't exist.
        prerelease_len = 0
        if not prefix:
            prefix = tuple()
        if isinstance(prefix, str):
            prefix = (prefix, )

        prerelease_len = len(prefix)

        current_prefix_match = starting_version.prerelease[:prerelease_len]

        if (new_version == starting_version.truncate() and
                prefix == current_prefix_match):

            # prerelease bumping for the same release.
            # this will catch a case where there's e.g. rc.dev.1
            # and we're trying to bump the 'rc' prefix, in that
            # case per semver, there is an implicit zero (i.e. .alpha comes
            # before .alpha.1).
            try:
                prerelease_version = int(
                    starting_version.prerelease[prerelease_len])

            except (IndexError, TypeError):
                # either there is no explicit prerelease version in the
                # current version, or there are additional prerelease
                # prefixes which are not intended for this bump
                prerelease_version = 0
        else:
            # different release or different prefix
            prerelease_version = 0

        prerelease_version += 1

        return prerelease_version

    def make_prerelease(self, version, prefix=None, build_date=None):
        """
        Make current version a prerelease

        Args:
            * version (`int` or `str`(digits only)): The numeric prerelease
                version. i.e., for -dev.3, this is `3` or `"3"`.
            * prefix (`str`, optional): the alphabetic (not enforced) prebuild
                prefix, such as `a`, `beta`, `dev`, and such.
            * build_date
        """

        prerelease_date = None
        time_fmt = "%Y%m%d%H%M%S"

        if build_date is True:
            prerelease_date = datetime.datetime.utcnow().strftime(time_fmt)

        elif build_date:
            prerelease_date = build_date.strftime(time_fmt)

        self.apply_prerelease((str(version)),
                              prefix=prefix,
                              build_date=prerelease_date)

    def apply_metadata(self, *metadata):
        """Apply build metadata to project version"""
        self._version.build += metadata

    def apply_prerelease(self, *prebuild, prefix=None, build_date=None):
        """Apply prebuild data to project version"""
        if prefix:
            self._version.prerelease = (prefix,)

        self._version.prerelease += tuple(str(element) for element in prebuild)

        if build_date is not None and build_date:

            self._version.prerelease += (build_date,)


def register_flavor(flavor):
    """
    Registers a Avakas Project Flavor
    Used for future language/project expansions
    """
    def wrapper(project):
        Avakas.project_flavors[flavor] = project
        return project
    return wrapper



#!/usr/bin/env python
import sys
import os


try:
    from setuptools import setup
except ImportError:
    print("avakas requires setuptools")
    sys.exit(1)


def main():
    vsn_path = "%s/version" % os.path.dirname(os.path.abspath(__file__))
    if not os.path.exists(vsn_path):
        print("%s is missing" % vsn_path)
        sys.exit(1)

    vsn_file = open(vsn_path, 'r')
    version = vsn_file.read()
    vsn_file.close()

    setup(name='avakas',
          version=version,
          description='Interact with project version metadata',
          author='Jonathan Freedman',
          author_email='jonafree@gmail.com',
          license='MIT',
          url='https://github.com/otakup0pe/avakas',
          install_requires=['semantic_version', 'gitpython', 'erl_terms'],
          packages=['avakas', 'avakas.flavors'],
          entry_points={
            'console_scripts': ['avakas = avakas.cli:main']
          },
          include_package_data=True,
          package_data={'avakas': [
            'version'
          ]})


if __name__ == "__main__":
    main()



"""
Avakas Built-In Chef Project Flavor
"""

import os
import re

from avakas.flavors.base import AvakasLegacy
from avakas.avakas import register_flavor
from avakas.errors import AvakasError
from avakas.utils import match_and_rewrite_lines


@register_flavor('chef')
class AvakasChefProject(AvakasLegacy):
    """
    Chef Cookbook Avakas Project Flavor
    """
    PROJECT_TYPE = 'chef'

    @classmethod
    def guess_flavor(cls, directory):
        return os.path.exists("%s/metadata.rb" % directory)

    def __read_metadata_file(self):
        handle = open("%s/metadata.rb" % self.directory, 'r')
        data = handle.read()
        return data, handle

    def read(self):
        """Extract the version from Chef Cookbook metadata"""
        data, handle = self.__read_metadata_file()
        pattern = r'^version.+["\'](?P<vsn>\d+\.\d+\.\d+)["\'].*'
        matcher = re.compile(pattern, re.MULTILINE)
        vsn_match = matcher.search(data)
        handle.close()
        self.version = str(vsn_match.group('vsn'))
        return True

    def write(self):
        """Writes the version to metadata.rb"""

        self.check_if_dirty()

        pattern = r'^(version.+["\'])(\d+\.\d+\.\d+)(["\'].*)'
        data, handle = self.__read_metadata_file()
        rewritten, updated = match_and_rewrite_lines(pattern, data,
                                                     str(self.version))

        if not updated:
            raise AvakasError('Unable to set version on metadata.rb')

        handle = open("%s/metadata.rb" % self.directory, 'w')
        handle.write(rewritten)
        handle.close()

        self.write_versionfile()

        self.write_git()



"""
Avakas Built-In Base Project Flavor
"""

import os
import re
import sys

from git import Repo

from avakas.errors import AvakasError
from avakas.avakas import Avakas, register_flavor
from avakas.utils import stdout_redirect


@register_flavor('legacy')
class AvakasLegacy(Avakas):
    """
    Default Legacy Avakas Project Flavor
    """
    PROJECT_TYPE = 'legacy'

    @classmethod
    def guess_flavor(cls, directory):
        """
        Return true if determined this is the project's flavor.
        For example, current directory has a meta/version file,
        return a true value.
        """
        os.path.exists(directory)
        # For legacy, this should _ALWAYS_ return False
        return False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.repo = None
        self.version_filename = kwargs['filename']
        self.commit_files = [self.version_filename]

    def __load_git(self):
        """Initializes our local git workspace."""
        opt = self.options
        repo = Repo(self.directory, search_parent_directories=True)
        if not repo:
            raise AvakasError("Unable to find associated git repo for %s." %
                              self.directory)

        if opt['branch'] not in repo.heads:
            raise AvakasError("Branch %s branch not found." % opt['branch'])

        if repo.active_branch != repo.heads[opt['branch']]:
            print("Switching to %s branch" % opt['branch'],
                  file=sys.stderr)
            repo.heads[opt['branch']].checkout()

        if opt['remote'] not in [r.name for r in repo.remotes]:
            raise AvakasError("Remote %s not found" % opt['remote'])

        # we really do not want to be polluting our stdout when
        # showing the version
        with stdout_redirect():
            repo.remotes[opt['remote']].pull(refspec=opt['branch'])

        return repo

    def __git_push(self, tag=None):
        """Push git commit or tag to remote"""
        opt = self.options
        if tag:
            resp = self.repo.remotes[opt['remote']].push(tag)

        resp = self.repo.remotes[opt['remote']].push()
        resp = resp[0]
        if resp.flags & 1024 or resp.flags & 32 or resp.flags & 16:
            raise AvakasError("Unexpected git error: %s" % resp.summary)

    def __commit_files(self):
        """Will commit and push the version file and optionally tags."""
        opt = self.options

        self.repo.index.add(self.commit_files)
        skip_hooks = not opt['with_hooks']
        self.repo.index.commit("Version bumped to %s" % self.version,
                               skip_hooks=skip_hooks)

    def __create_git_tag(self):
        """Creates a git tag"""
        tag = self.version
        self.repo.create_tag(tag)

        return tag

    def __determine_bump(self):
        """Will go through the Git history until the last version bump
        and look for hints that we want to "automatically" bump
        our version"""
        self.repo = self.__load_git()
        vsn = None
        reg = re.compile(r'(\#|bump:|\[)(?P<bump>(patch|minor|major))(.*|\])',
                         re.MULTILINE)
        for commit in self.repo.iter_commits(self.options['branch']):
            # we go iterate back to the last time we bumped the version
            if commit.message.startswith('Version bumped to'):
                break

            res = reg.search(commit.message)
            if res:
                bump = res.group('bump')
                if not vsn:
                    vsn = bump
                elif vsn == 'patch' and bump == 'minor':
                    vsn = 'minor'
                elif vsn == 'patch' and bump == 'major':
                    vsn = 'major'
                elif vsn == 'minor' and bump == 'major':
                    vsn = 'major'
        return vsn

    def check_if_dirty(self):
        """Check if the repo is dirty"""
        self.repo = self.__load_git()
        if not self.options['skipdirty'] and self.repo.is_dirty():
            raise AvakasError("Git repo dirty.")

    def write_versionfile(self):
        """Write the version file"""

        path = os.path.join(self.directory, self.version_filename)
        version_file = open(path, 'w')
        version_file.write("%s\n" % self.version)
        version_file.close()

    def write_git(self):
        """Write data to git"""
        tag = None

        if not self.options['dry']:
            if self.version_filename and self.options['commitchanges']:
                self.__commit_files()
                self.__git_push()

            if not self._version.build:
                tag = self.__create_git_tag()
                self.__git_push(tag=tag)

    def bump(self,
             bump=None,
             prerelease=False,
             prerelease_prefix=None, build_date=None):
        """
        When using 'auto', flavor will attempt to determine whether or not
        the project needs to be bumped from git log history. If keywords are
        detected, project will update it's version to detected bump level and
        return True. If no keywords are detected, default_bump level will be
        used and True returned. If not, bump will return False.
        """
        if bump == 'auto':
            bump = self.__determine_bump()
            if bump is None and self.options['default_bump']:
                bump = self.options['default_bump']

        return super().bump(
            bump=bump,
            prerelease=prerelease,
            prerelease_prefix=prerelease_prefix,
            build_date=build_date)

    def read(self):
        """
        Get the version from the current project flavor
        """
        path = os.path.join(self.directory, self.version_filename)
        version_file = open(path, 'r')
        version_str = version_file.read()
        self.version = version_str
        version_file.close()
        return True

    def write(self):
        """
        Write version out to file
        """

        self.check_if_dirty()
        self.write_versionfile()
        self.write_git()
